/*
 * Copyright 2021 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.android.fhir.datacapture.contrib.views.barcode.mlkit.md.barcodedetection

import android.graphics.Canvas
import android.graphics.Color
import android.graphics.CornerPathEffect
import android.graphics.Paint
import android.graphics.Paint.Style
import android.graphics.PorterDuff
import android.graphics.PorterDuffXfermode
import android.graphics.RectF
import androidx.core.content.ContextCompat
import com.google.android.fhir.datacapture.contrib.views.barcode.R
import com.google.android.fhir.datacapture.contrib.views.barcode.mlkit.md.camera.GraphicOverlay
import com.google.android.fhir.datacapture.contrib.views.barcode.mlkit.md.camera.GraphicOverlay.Graphic
import com.google.android.fhir.datacapture.contrib.views.barcode.mlkit.md.settings.PreferenceUtils

/**
 * See the barcode detection classes here
 * https://github.com/googlesamples/mlkit/tree/master/android/material-showcase/app/src/main/java/com/google/mlkit/md/barcodedetection
 */
internal abstract class BarcodeGraphicBase(overlay: GraphicOverlay) : Graphic(overlay) {

  private val boxPaint: Paint =
    Paint().apply {
      color = ContextCompat.getColor(context, R.color.barcode_reticle_stroke)
      style = Style.STROKE
      strokeWidth =
        context.resources.getDimensionPixelOffset(R.dimen.barcode_reticle_stroke_width).toFloat()
    }

  private val scrimPaint: Paint =
    Paint().apply { color = ContextCompat.getColor(context, R.color.barcode_reticle_background) }

  private val eraserPaint: Paint =
    Paint().apply {
      strokeWidth = boxPaint.strokeWidth
      xfermode = PorterDuffXfermode(PorterDuff.Mode.CLEAR)
    }

  val boxCornerRadius: Float =
    context.resources.getDimensionPixelOffset(R.dimen.barcode_reticle_corner_radius).toFloat()

  val pathPaint: Paint =
    Paint().apply {
      color = Color.WHITE
      style = Style.STROKE
      strokeWidth = boxPaint.strokeWidth
      pathEffect = CornerPathEffect(boxCornerRadius)
    }

  val boxRect: RectF = PreferenceUtils.getBarcodeReticleBox(overlay)

  override fun draw(canvas: Canvas) {
    // Draws the dark background scrim and leaves the box area clear.
    canvas.drawRect(0f, 0f, canvas.width.toFloat(), canvas.height.toFloat(), scrimPaint)
    // As the stroke is always centered, so erase twice with FILL and STROKE respectively to clear
    // all area that the box rect would occupy.
    eraserPaint.style = Style.FILL
    canvas.drawRoundRect(boxRect, boxCornerRadius, boxCornerRadius, eraserPaint)
    eraserPaint.style = Style.STROKE
    canvas.drawRoundRect(boxRect, boxCornerRadius, boxCornerRadius, eraserPaint)
    // Draws the box.
    canvas.drawRoundRect(boxRect, boxCornerRadius, boxCornerRadius, boxPaint)
  }
}
